#!/usr/bin/env python

"""Web Crawler/Spider

This module implements a web crawler. This is very _basic_ only
and needs to be extended to do anything usefull with the
traversed pages.

From code.activestate.com/recipes/576551-simple-web-crawler/

"""

import re
"""
Basic dfns::
  -- crawler:  given a starting page C{root} crawls over the page collecting content.  Follows links
  FROM page (possibly only site-external links, possibly only site-internal, possibly both),
  crawls link targets and repeats.  Up to termination or some designated depth.
  -- C{fetcher}:  Fetches some designated info from a single page.  Possibly content.
  Possibly links.  Possibly metainfo, and so on.
  -- This crawler is created with an optional start url (C{root}), but does not crawl until told to
  do so with the C{crawl} method.  Strangely enough another url may NOT be passed in to C{crawl}
  to supercede the creation-time C{root} url. Possible change. But at the moment a single crawler instance
  is rarely used with more than one root.  Crawler crawls by creating a new fetcher for every
  call to C{crawl}.
  -- This fetcher always collects links and optionally collects page content.  The fetcher knows
  nothing about internal or external links and makes no decisions about what to follow next.
  That is the crawler's domain (control structure).  Therefore teh fether just blindly collects
  links, not looking at them.
  -- After the fetcher returns with info from a single page, crawler scrapes it to see if
  there are new URLs to crawl.  If there are and other conditions are met, the crawler crawls them.
  -- All links followed are stored on the crawler in the C{followed_links} attribute.
"""

import sys, os.path
import time
import math
import urllib2
import urlparse
import optparse
import codecs
#  Keep module name intact
import BeautifulSoup
# Finally import clean up module.
import crawler_package.crawling_utils

__version__ = "0.1"
__copyright__ = "CopyRight (C) 2008 by James Mills"
__license__ = "GPL"
__author__ = "James Mills"
__author_email__ = "James Mills, James dot Mills st dotred dot com dot au"

USAGE = "%prog [options] <url>"
VERSION = "%prog v" + __version__

AGENT = "%s/%s" % (__name__, __version__)

class Queue(object):

    def __init__(self, size=None):
        super(Queue, self).__init__()

        self._queue = []
        self._size = size

    def __len__(self):
        return len(self._queue)

    def __getitem__(self, n):
        if (not self.empty()) and (0 <= (n + 1) <= len(self._queue)):
            return self._queue[(len(self._queue) - (n + 1))]
        else:
            raise StopIteration

    def push(self, item):
        self._queue.insert(0, item)
        if self._size is not None:
            self._queue = self._queue[:self._size]

    def get(self, n=0, remove=False):
        if (not self.empty()) and (0 <= (n + 1) <= len(self._queue)):
            r = self._queue[(len(self._queue) - (n + 1))]
            if remove:
                del self._queue[(len(self._queue) - (n + 1))]
            return r
        else:
            return None

    def pop(self, n=0):
        return self.get(n, True)

    def peek(self, n=0):
        return self.get(n)

    def top(self):
        return self.peek()

    def bottom(self):
        return self.peek(len(self) - 1)

    def empty(self):
        return self._queue == []

    def size(self):
        return self._size

    def full(self):
        return len(self) == self.size()

class Crawler(object):

    def __init__(self, root, depth, locked=True, write_dir = None):
        self.root = root
        self.depth = depth
        self.locked = locked
        self.host = urlparse.urlparse(root)[1]
        self.links = 0
        self.followed = 0
        self.contents = {}
        self.raw_contents = {}
        self.meta_info = {}
        if not write_dir:
              self._write_dir_ = os.getcwd()
        else:
            self._write_dir_ = write_dir

    def crawl(self, store_content=False, write_content=False):
        """
        C{write_content} = True presupposes C{store_content}=True.
        """
        self.current_host = crawler_package.crawling_utils.simplify_netloc(self.host)
        page = self.call_fetcher(self.root, store_content, write_content)
        urls = Queue()
        for url in page.urls:
            urls.push(url)
        followed = [self.root]
        self.followed_links = followed
        external = []
        self.external = external

        n = 0

        while not urls.empty():
            n += 1
            url = urls.pop()
            if url not in followed:
                try:
                    host = crawler_package.crawling_utils.simplify_netloc(urlparse.urlparse(url)[1])
                    if host and self.current_host:
                        ## Allow current hostname to have prefix.  Why?
                        internal_link = re.match(".*%s" % self.current_host, host)
                    elif host:
                        # Therfore self.current_host is the empty string. No match.
                        internal_link = False
                        #internal_link = True
                    else:
                        # Both lack hosts.  Therefore this link is internal
                        internal_link= True
                    if not internal_link:
                        external.append(url)
                    if self.locked and internal_link:
                        followed.append(url)
                        self.followed += 1
                        page = self.call_fetcher(url, \
                                                 store_content,\
                                                 write_content)
                        for i, url in enumerate(page):
                            if url not in urls:
                                self.links += 1
                                urls.push(url)
                        if n > self.depth and self.depth > 0:
                            break
                except Exception, error:
                    print "Warning: Can't process url '%s'" % url
                    print error

    def call_fetcher (self, url, store_content, write_content):
        """
        C{write_content} = True means write content to a
        file now.
        
        Writing content incrementally for each page highly
        recommended because of the high likelihood of
        web hangups while following all the links on a crawl.
        """
        page = Fetcher(url)
        page.fetch(store_content=store_content)
        if store_content and page.soup is not None:
            content = page.pretty_content
            self.contents[url] = content
            self.raw_contents[url] = page.soup
            self.meta_info[url] = self.get_keywords_description(page.soup)
            if write_content:
                print 'writing content'
                full_path = crawler_package.crawling_utils.write_website_file(self._write_dir_,\
                                                              url,\
                                                              content, \
                                                              clean=True)
                self.print_meta_info(url, full_path)
                print 'written'
        return page

    def get_content (self, index=None, url=None):
        """
        Just a method for retrieving some
        content that has previously been stored,
        either by url, or, somewhat arbitrarily,
        by an index that will be used on
        the keylist of the C{contents } dictionary.
        """
        assert not index and url
        assert index or url
        contents = self.contents
        if index is not None:
            try:
                keys = contents.keys()
                url = keys[index]
            except IndexError:
                if index == 0:
                    print 'There were no URLs'
                else:
                    print 'There are only %d URLs.' % len(keys)
        try:
            return contents[url]
        except KeyError:
            print 'No such URL has been crawled'

    def get_keywords_description(self,soup):
        """
        Extract name and content from meta tags.
        For example, name may be keywords or description,
        crucial info about a site.
        """
        meta_list = soup.findAll('meta')
        content_dict = { }
        for tag in  meta_list:
            try:
                name = tag['name']
                content = tag['content']
            except:
                continue
            content_dict[name] = content
        return content_dict
        

    def dump_content(self):
        contents = self.contents
        for url in contents:
            full_path = crawler_package.crawling_utils.write_content_file(self._write_dir_, \
                                                          url, \
                                                          contents[url])
            self.print_meta_info(url, full_path)


    def print_meta_info(self,url, full_path):
        meta_info = self.meta_info[url]
        soup = self.raw_contents[url] 
        meta_info['html'] = soup
        for k in meta_info:
            try:
                fh = open(full_path + os.path.extsep + k, 'w')
            except IOError, message:
                print 'IOError', message
            print >> fh, '<html>'
            print >> fh, '<p>'
            if type(meta_info[k]) == type(u''):
                print >> fh, codecs.utf_8_encode(meta_info[k],'ignore')
            else:
                print >> fh, meta_info[k]
            fh.close()

    def print_content_index(self):
        keys = self.contents.keys()
        for index in range(len(keys)):
            print index, keys[index]


class Fetcher(object):

    def __init__(self, url):
        self.url = url
        self.urls = []
        self.pretty_content = []
        self.soup = None

    def __contains__(self, x):
        return x in self.urls

    def __getitem__(self, x):
        return self.urls[x]

    def _addHeaders(self, request):
        request.add_header("User-Agent", AGENT)

    def open(self):
        url = self.url
        print "Following %s" % url
        try:
            request = urllib2.Request(url)
            handle = urllib2.build_opener()
        except IOError:
            return None
        return (request, handle)

    def store_content(self, soup):
        pretty_content_list = self.pretty_content
        crawler_package.crawling_utils.collect_text_lines(pretty_content_list,[soup])
        # self.pretty_content = pretty_content_list

    def fetch(self, store_content=False):
        request, handle = self.open()
        # print request, handle
        self._addHeaders(request)
        if handle:
            soup = BeautifulSoup.BeautifulSoup()
            self.soup = soup
            try:
                content = unicode(handle.open(request).read(), errors="ignore")
                soup.feed(content)
                if store_content:
                    self.store_content(soup)
                tags = soup('a')
                alt_tags = soup('select')
            except urllib2.HTTPError, error:
                if error.code == 404:
                    print >> sys.stderr, "ERROR: %s -> %s" % (error, error.url)
                else:
                    print >> sys.stderr, "ERROR: %s" % error
                tags = []
            except urllib2.URLError, error:
                print >> sys.stderr, "ERROR: %s" % error
                tags = []
            for tag in tags:
                try:
                    href = tag["href"]
                    if href is not None:
                        url = urlparse.urljoin(self.url, crawler_package.crawling_utils.encodeHTML(href))
                        if url not in self:
                            # print " Found: %s" % url
                            self.urls.append(url)
                except KeyError:
                    pass
            ## JMG 7/13/11.  Added this OTHER way of finding links
            for tag in alt_tags:
                options = tag('option')
                for o in options:
                    try:
                        value = o['value']
                        if value:
                            (scheme, netloc, path, params, query, fragment) = urlparse.urlparse(value)
                            if netloc and value not in self:
                                self.urls.append(value)
                    except KeyError:
                        pass
            # print 'fetcher: %s' % self.urls
                        


def getLinks(url):
    page = Fetcher(url)
    page.fetch()
    for i, url in enumerate(page):
        print "%d. %s" % (i, url)


def findLinks(url):
    urlparser = urllib2.urlparse.urlparse
    (scheme, root_host, path, params, query, fragment) = urlparser(url)
    page = Fetcher(url)
    page.fetch()
    links = set([])
    for i, lurl in enumerate(page):
        link_host = urlparser(lurl)[1]
        if link_host <> root_host:
            links.add((link_host,lurl))
    return links

def parse_options(url=None):
    """parse_options() -> opts, args

    Parse any command-line options given returning both
    the parsed options and arguments.
    """

    parser = optparse.OptionParser(usage=USAGE, version=VERSION)

    parser.add_option("-q", "--quiet",
            action="store_true", default=False, dest="quiet",
            help="Enable quiet mode")

    parser.add_option("-l", "--links",
            action="store_true", default=False, dest="links",
            help="Get links for specified url only")

    parser.add_option("-d", "--depth",
            action="store", type="int", default=260, dest="depth",
            help="Maximum depth to traverse")

    parser.add_option("-f", "--fetch_only",
            action="store_true", default=False, dest="fetch_only",
            help="Fetch root page only, no links followed")

    opts, args = parser.parse_args()

    if len(args) < 1 and not url:
        parser.print_help()
        raise SystemExit, 1

    return opts, args


def main(url=None,write_dir=None, store_content=True,\
         write_content=True, depth=None, fetch_only=False):
    """
    Interface for using crawler as a script.

    """

    global crawler
    ## Shd be sys.args?
    opts, args = parse_options(url)

    if not url:
        url = args[0]

    if opts.links:
        getLinks(url)
        raise SystemExit, 0

    if depth is None:
        depth = opts.depth

    if opts.fetch_only:
        fetch_only = True

    assert not fetch_only and depth, 'Cant specify depth in fetch_only mode!'

    sTime = time.time()

    crawler = Crawler(url, depth,write_dir=write_dir)
    call_crawler(crawler, url, write_dir=write_dir,\
                 store_content=store_content,write_content=write_content,\
                 fetch_only=fetch_only,depth=depth)

    eTime = time.time()
    tTime = eTime - sTime

    print "Found:    %d" % crawler.links
    print "Followed: %d" % crawler.followed
    print "Stats:    (%d/s after %0.2fs)" % (
            int(math.ceil(float(crawler.links) / tTime)), tTime)

def  call_crawler(crawler, url, write_dir=None,\
                  store_content=True,write_content=True,
                  fetch_only=False,depth=None):
    """
    Provides an interface for external programs calling the crawler.
    """
    
    if fetch_only:
        print "Fetching %s" % (url,)
        crawler.call_fetcher(url,store_content, write_content)
    else:
        print "Crawling %s (Max Depth: %d)" % (url, depth)
        print 'URL: %s' % (url,)
        crawler.crawl(store_content=store_content, write_content=write_content)
        
    if not write_content:
        #  If not writing it incrementally, dump it all now.
        crawler.dump_content()


if __name__ == "__main__":
    ### remote website access
    root_url = "http://www.splcenter.org/get-informed/intelligence-files/groups"
    ###  Local file access
    home = os.environ['HOME']
    args = sys.argv[1: ]
    file_url = None
    if len(args) < 1:
        print 'Usage %s <website_file> [write_dir]' % (sys.argv[0],)
        # sys.exit()
    else:
        file_url = args[0]
        args = args[1:]
    if args:
        new_data_dir = args[0]
    else:
        new_data_dir = os.getcwd()
    print 'Using %s as websites file and %s as write directory' % (file_url, new_data_dir)
    root_dir = os.path.join(home, 'Desktop', 'SWIDSAS')
    working_dir = os.path.join(root_dir,'software','scripts')
    file_name = 'test.html'
    file_path= os.path.join(working_dir, file_name)
    file_url = 'file://' + file_path
    ###  Pick a root to crawl from
    # main(url=root_url)
    # file_url='http://whitereference.blogspot.com/2010_05_16_archive.html'
    main(url=file_url,write_dir=new_data_dir)
    # To look into the soup for a random page.
    # soup = crawler.raw_contents.values()[0]
